import logging
import os
from typing import Any, Literal

from lettrade.account import Account
from lettrade.brain import Brain
from lettrade.commander import Commander
from lettrade.data import DataFeed, DataFeeder
from lettrade.exchange import Exchange
from lettrade.plot import Plotter
from lettrade.stats import BotStatistic
from lettrade.strategy import Strategy

logger = logging.getLogger(__name__)


class LetTradeBot:
    """Bot object helps to manage actions of bot"""

    datas: list[DataFeed]
    """DataFeed list for bot"""
    data: DataFeed
    """Main DataFeed for bot"""

    brain: Brain
    """Brain of bot"""
    feeder: DataFeeder
    """DataFeeder help to handle `datas`"""
    exchange: Exchange
    """Trading exchange and events"""
    account: Account
    """Trading account handler"""
    strategy: Strategy
    """Strategy"""
    commander: Commander | None = None
    """Control the bot"""
    plotter: Plotter | None = None
    """Plot graphic results"""
    stats: BotStatistic | None = None

    _strategy_cls: type[Strategy]
    _feeder_cls: type[DataFeeder]
    _exchange_cls: type[Exchange]
    _account_cls: type[Account]
    _commander_cls: type[Commander] | Commander | None
    _plotter_cls: type[Plotter] | None
    _stats_cls: type[BotStatistic] | None
    _kwargs: dict[str, Any]
    _name: str | None

    def __init__(
        self,
        datas: DataFeed | list[DataFeed] | list[str] | str,
        strategy: type[Strategy],
        feeder: type[DataFeeder],
        exchange: type[Exchange],
        account: type[Account],
        commander: type[Commander] | None = None,
        plotter: type[Plotter] | None = None,
        stats: type[BotStatistic] | None = None,
        name: str | None = None,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            datas (DataFeed | list[DataFeed] | list[str] | str): _description_
            strategy (type[Strategy]): _description_
            feeder (type[DataFeeder]): _description_
            exchange (type[Exchange]): _description_
            account (type[Account]): _description_
            commander (type[Commander] | None, optional): _description_. Defaults to None.
            plotter (type[Plotter] | None, optional): _description_. Defaults to None.
            stats (type[BotStatistic] | None, optional): _description_. Defaults to None.
            name (str | None, optional): _description_. Defaults to None.
        """
        logger.info("New bot: %s", name)

        self._strategy_cls = strategy
        self._feeder_cls = feeder
        self._exchange_cls = exchange
        self._account_cls = account
        self._commander_cls = commander
        self._plotter_cls = plotter
        self._stats_cls = stats

        self._name = name
        self._kwargs = kwargs

        # DataFeeds
        self.datas = datas
        self.data = self.datas[0]

    def init(self):
        """Init objects from classes"""
        # Feeder
        self.feeder = self._feeder_cls(**self._kwargs.get("feeder_kwargs", {}))
        self.feeder.init(self.datas)

        # Account
        self.account = self._account_cls(**self._kwargs.get("account_kwargs", {}))

        # Exchange
        self.exchange = self._exchange_cls(**self._kwargs.get("exchange_kwargs", {}))

        # Commander
        if self._commander_cls:
            if isinstance(self._commander_cls, Commander):
                self.commander = self._commander_cls
            else:
                self.commander = self._commander_cls(
                    **self._kwargs.get("commander_kwargs", {})
                )

        # Strategy
        self.strategy = self._strategy_cls(
            feeder=self.feeder,
            exchange=self.exchange,
            account=self.account,
            commander=self.commander,
            **self._kwargs.get("strategy_kwargs", {}),
        )

        # Brain
        self.brain = Brain(
            strategy=self.strategy,
            exchange=self.exchange,
            feeder=self.feeder,
            commander=self.commander,
            **self._kwargs.get("brain_kwargs", {}),
        )

        # Init
        if self.commander:
            self.commander.init(
                bot=self,
                brain=self.brain,
                exchange=self.exchange,
                strategy=self.strategy,
            )
        self.exchange.init(
            brain=self.brain,
            feeder=self.feeder,
            account=self.account,
            commander=self.commander,
        )

        # Stats
        self.stats = self._stats_cls(
            feeder=self.feeder,
            exchange=self.exchange,
            strategy=self.strategy,
            **self._kwargs.get("stats_kwargs", {}),
        )

        # Plotter
        if self._plotter_cls:
            self.plotter = self._plotter_cls(
                self,
                **self._kwargs.get("plotter_kwargs", {}),
            )

        if __debug__:
            logger.debug("Bot %s inited with %d datas", self._name, len(self.datas))

    def start(self, **kwargs):
        """Start bot"""
        if not hasattr(self, "brain"):
            self.init()

        self.brain.start(**kwargs)

        if __debug__:
            logger.debug("Bot %s started with %d datas", self._name, len(self.datas))

    def run(self):
        """Run bot"""
        if self.commander:
            self.commander.start()

        self.brain.run()

        if self.commander:
            self.commander.stop()

        if self._stats_cls:
            self.stats.compute()

    def plot(self, *args, **kwargs):
        """Plot bot result"""
        if __debug__:
            from .utils.docs import is_docs_session

            if is_docs_session():
                return

        self.plotter.plot(*args, **kwargs)

    def stop(self):
        """Stop bot"""
        self.brain.stop()
        if self.plotter is not None:
            self.plotter.stop()

    @classmethod
    def new_bot(
        cls,
        datas: list[DataFeed],
        strategy_cls: type[Strategy],
        feeder_cls: type[DataFeeder],
        exchange_cls: type[Exchange],
        account_cls: type[Account],
        commander_cls: type[Commander],
        plotter_cls: type[Plotter],
        stats_cls: type[BotStatistic],
        name: str | None = None,
        **kwargs,
    ) -> "LetTradeBot":
        """Create new bot object

        Args:
            datas (list[DataFeed]): _description_
            strategy_cls (type[Strategy]): _description_
            feeder_cls (type[DataFeeder]): _description_
            exchange_cls (type[Exchange]): _description_
            account_cls (type[Account]): _description_
            commander_cls (type[Commander]): _description_
            plotter_cls (type[Plotter]): _description_
            stats_cls (type[BotStatistic]): _description_
            name (str | None, optional): _description_. Defaults to None.

        Returns:
            LetTradeBot: _description_
        """
        bot = cls(
            strategy=strategy_cls,
            datas=datas,
            feeder=feeder_cls,
            exchange=exchange_cls,
            account=account_cls,
            commander=commander_cls,
            plotter=plotter_cls,
            stats=stats_cls,
            name=name,
            **kwargs,
        )
        return bot

    @classmethod
    def start_bot(
        cls,
        bot: "LetTradeBot | None" = None,
        **kwargs,
    ) -> "LetTradeBot":
        """Init and start bot object

        Args:
            bot (LetTradeBot | None, optional): _description_. Defaults to None.

        Returns:
            LetTradeBot: _description_
        """
        if bot is None:
            bot = cls.new_bot(**kwargs)
        bot.init()
        bot.start(**kwargs.get("start_kwargs", {}))
        return bot

    @classmethod
    def run_bot(
        cls,
        bot: "LetTradeBot | None" = None,
        datas: list[DataFeed] | None = None,
        id: int | None = None,
        name: str | None = None,
        result: Literal["str", "stats", "bot", None] = "str",
        **kwargs,
    ) -> "LetTradeBot | BotStatistic | str | None":
        """Run bot object

        Args:
            bot (LetTradeBot | None, optional): _description_. Defaults to None.
            datas (list[DataFeed] | None, optional): _description_. Defaults to None.
            id (int | None, optional): _description_. Defaults to None.
            name (str | None, optional): _description_. Defaults to None.
            result (Literal["str", "stats", "bot", None], optional): _description_. Defaults to "str".

        Returns:
            LetTradeBot | BotStatistic | str | None: Return value will be pickle across multiprocessing.
                The cost higher from `str`, `stats` object, `bot` object
        """
        # Set name for current processing
        if name is None:
            d = datas[0] if datas else bot.data
            name = f"{id}-{os.getpid()}-{d.name}"

        if bot is None:
            bot = cls.start_bot(
                datas=datas,
                name=name,
                **kwargs,
            )

        # bot
        bot.run(**kwargs.get("run_kwargs", {}))

        # Return type
        if result == "stats":
            return bot.stats
        if result == "str":
            return str(bot.stats)

        return bot



from typing import Literal

import pandas as pd

from ..series import series_init
from ..utils import talib_ma


def sma(
    series: pd.Series | str = "close",
    window: int = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """Simple Moving Average

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to None.
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list[str], optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Example:
        ```python
        df.i.sma(window=21, name="sma", inplace=True, plot=True)
        ```

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return ma(
        series=series,
        window=window,
        mode="sma",
        dataframe=dataframe,
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def ema(
    series: pd.Series | str = "close",
    window: int = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """Exponential Moving Average

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to None.
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Example:
        ```python
        df.i.ema(window=21, name="ema", inplace=True, plot=True)
        ```

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return ma(
        series=series,
        window=window,
        mode="ema",
        dataframe=dataframe,
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def ma(
    series: pd.Series | str = "close",
    window: int = None,
    mode: Literal[
        "sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3"
    ] = None,
    dataframe: pd.DataFrame = None,
    name: str | None = None,
    prefix: str = "",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """Moving Average

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to None.
        mode (Literal[ "sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3" ], optional): _description_. Defaults to None.
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.
        plot (bool | list, optional): _description_. Defaults to False.
        plot_kwargs (dict | None, optional): _description_. Defaults to None.

    Example:
        ```python
        df.i.ma(window=21, name="sma", mode="sma", inplace=True, plot=True)
        ```

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    # Validation & init
    if __debug__:
        if window is None or window <= 0:
            raise RuntimeError(f"Window {window} is invalid")
        if mode is None:
            raise RuntimeError(f"Mode {window} is invalid")
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    series = series_init(series=series, dataframe=dataframe, inplace=inplace)

    # Indicator
    ma_fn = talib_ma(mode)
    i = ma_fn(series, timeperiod=window, **kwargs)

    if inplace:
        name = name or f"{prefix}{mode}"
        dataframe[name] = i

        # Plot
        if plot:
            if plot_kwargs is None:
                plot_kwargs = dict()

            plot_kwargs.update(series=name, name=name)

            from lettrade.indicator.plot import IndicatorPlotter
            from lettrade.plot.plotly import plot_line

            IndicatorPlotter(dataframe=dataframe, plotter=plot_line, **plot_kwargs)

        return dataframe

    return i



from .atr import atr
from .bollinger_bands import bollinger_bands
from .keltner_channel import keltner_channel


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    from ..series import series_indicator_inject

    obj.bollinger_bands = series_indicator_inject(bollinger_bands)

    obj.keltner_channel = keltner_channel
    obj.atr = atr



import pandas as pd


def resample(df: pd.DataFrame, timeframe: str = "5min") -> pd.DataFrame:
    ohlc_dict = {
        "open": "first",
        "high": "max",
        "low": "min",
        "close": "last",
        "volume": "sum",
    }
    # Resample to "left" border as dates are candle open dates
    df = df.resample(timeframe, label="left").agg(ohlc_dict).dropna()
    return df



from typing import TYPE_CHECKING

from lettrade.account import Account

if TYPE_CHECKING:
    from lettrade.exchange.backtest import BackTestPosition


class BackTestAccount(Account):
    _commission: float

    def __init__(
        self,
        risk: float = 0.02,
        balance: float = 10_000,
        commission: float = 0.2,
        margin: float = 1,
        leverage: float = 1,
        **kwargs,
    ) -> None:
        """Account for backtest

        Args:
            risk (float, optional): _description_. Defaults to 0.02.
            balance (float, optional): _description_. Defaults to 10_000.
            commission (float, optional): Commission fee is percent of size. Defaults to 0.2.
            margin (float, optional): _description_. Defaults to 1.
            leverage (float, optional): _description_. Defaults to 1.
            **kwargs (dict, optional): Mirror of [lettrade.account.Account()](site:/reference/account/account/#lettrade.account.account.Account).
        """
        super().__init__(
            risk=risk,
            balance=balance,
            margin=margin,
            leverage=leverage,
            **kwargs,
        )
        self._commission = commission

    def __repr__(self):
        return "<BackTestAccount " + str(self) + ">"

    @property
    def equity(self) -> float:
        equity = self._balance
        if len(self._exchange.positions) > 0:
            equity += sum(position.pl for position in self._exchange.positions.values())
        return equity

    def pl(self, size, entry_price: float, exit_price: float | None = None) -> float:
        if exit_price is None:
            exit_price = self._exchange.data.l.open[0]

        pl = size * (exit_price - entry_price)

        return pl

    def fee(self, size: float, **kwargs: dict):
        return -abs(size * self._commission)

    def on_positions(self, positions: list["BackTestPosition"]):
        for position in positions:
            if position.is_exited:
                self._balance += position.pl
        super().on_positions(positions)


class ForexBackTestAccount(BackTestAccount):
    """Forex backtest account helps to handle lot size"""

    def __repr__(self):
        return "<ForexBackTestAccount " + str(self) + ">"

    def pl(self, size, entry_price: float, exit_price: float | None = None) -> float:
        return super().pl(size, entry_price, exit_price) * 100_000



import unittest

import pandas as pd

from lettrade.data import DataFeed
from lettrade.exchange.backtest.data import CSVBackTestDataFeed


class DataFeedInjectTestCase(unittest.TestCase):
    def setUp(self):
        self.data: DataFeed = CSVBackTestDataFeed("test/assets/EURUSD_1h-0_1000.csv")
        self.data._set_main()

    def test_pointer(self):
        df = self.data.copy(deep=True)

        self.assertEqual(df.l[0].name, df.l.index[0], "Data[0] index value wrong")
        self.assertEqual(df.l[0].open, df.l.open[0], "Data[0] open value wrong")
        self.assertEqual(df.l[0].high, df.l.high[0], "Data[0] high value wrong")
        self.assertEqual(df.l[0].low, df.l.low[0], "Data[0] low value wrong")
        self.assertEqual(df.l[0].close, df.l.close[0], "Data[0] close value wrong")
        self.assertEqual(df.l[0].volume, df.l.volume[0], "Data[0] volume value wrong")

        self.assertIs(
            df.l._pointer,
            df.l.index.pointer,
            "Data.l._pointer value wrong",
        )
        self.assertIs(
            df.l._pointer,
            df.l.open.pointer,
            "Data.l._pointer value wrong",
        )
        self.assertIs(
            df.l._pointer,
            df.l.high.pointer,
            "Data.l._pointer value wrong",
        )
        self.assertIs(
            df.l._pointer,
            df.l.low.pointer,
            "Data.l._pointer value wrong",
        )
        self.assertIs(
            df.l._pointer,
            df.l.close.pointer,
            "Data.l._pointer value wrong",
        )
        self.assertIs(
            df.l._pointer,
            df.l.volume.pointer,
            "Data.l._pointer value wrong",
        )

    def test_next(self):
        df = self.data.copy(deep=True)

        df.next()
        self.assertEqual(
            df.l.pointer,
            df.l.index.pointer,
            "Data.l.pointer value wrong",
        )
        self.assertEqual(
            df.open.iloc[1],
            df.l.open.iloc[0],
            "Data..l.open.iloc[0] value wrong",
        )

    def test_overrite_index(self):
        df = self.data.copy(deep=True)

        index_id = id(df.index)
        pointer_id = id(df.l.pointer)

        dt = pd.Timestamp("2024-01-01 00:00:00")

        self.assertTrue(df[df.index == dt].empty, f"Datetime {dt} existed in data")

        df.loc[dt, ["open", "close"]] = [1, 2]

        self.assertNotEqual(index_id, id(df.index))
        self.assertEqual(pointer_id, id(df.l.index.pointer))

    def test_set_index(self):
        df = self.data.copy(deep=True)

        index_id = id(df.index)
        pointer_id = id(df.l.pointer)

        df.reset_index(inplace=True)
        df.set_index("datetime", inplace=True)

        self.assertNotEqual(index_id, id(df.index))
        self.assertEqual(pointer_id, id(df.l.index.pointer))


if __name__ == "__main__":
    unittest.main(verbosity=2)



if __name__ == "__main__":
    from setuptools import find_packages, setup

    with open("README.md", encoding="utf-8") as fh:
        long_description = fh.read()

    setup(
        name="lettrade",
        version="0.0.10-beta-4",
        author="Santatic",
        # author_email = "author@example.com",
        description="Lightweight trading framwork",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/AwesomeTrading/LetTrade",
        classifiers=[
            "Intended Audience :: Financial and Insurance Industry",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Topic :: Office/Business :: Financial :: Investment",
            "Topic :: Scientific/Engineering :: Visualization",
        ],
        packages=find_packages(),
        python_requires=">=3.10,<3.13",
        install_requires=["pandas>=2.0", "numpy<2.0", "numexpr", "TA-Lib>=0.4"],
        extras_require={
            "plot": ["plotly"],
            "jupyter": ["nbformat", "rich[jupyter]"],
            "commander": ["lettrade[plot]"],
            "commander-telegram": ["lettrade[commander]", "python-telegram-bot"],
            "backtest": ["lettrade[plot]", "rich"],
            "backtest-extra": ["lettrade[backtest]", "lettrade[jupyter]", "yfinance"],
            "live": ["lettrade[plot]"],
            "exchange-metatrader": [
                "lettrade[live]",
                "python-box",
                "mt5linux",
                # "mt5linux @ git+https://github.com/AwesomeTrading/mt5linux.git@master"
            ],
            "exchange-ccxt": ["lettrade[live]", "ccxt", "python-box"],
            "test": ["pytest"],
            "all": [
                "lettrade[backtest-extra]",
                "lettrade[exchange-metatrader]",
                "lettrade[exchange-ccxt]",
                "lettrade[commander-telegram]",
                "lettrade[test]",
            ],
        },
    )



# import unittest

# import numpy as np
# import pandas as pd
# import talib.abstract as ta
# from pandas import testing as pdtest

# from lettrade.exchange.backtest.data import CSVBackTestDataFeed
# from lettrade.indicator.vendor.qtpylib import qtpylib


# class QTPyLibIndicatorTestCase(unittest.TestCase):
#     def setUp(self):
#         self.data = CSVBackTestDataFeed("test/assets/EURUSD_1h-0_1000.csv")
#         self.data._set_main()

#         self.raw_data = pd.read_csv(
#             "test/assets/EURUSD_1h-0_1000.csv",
#             index_col=0,
#             parse_dates=["datetime"],
#         )

#         # Ta-Lib
#         self.ta_ema = ta.EMA(self.data, timeperiod=21)
#         self.ta_ema_raw = ta.EMA(self.raw_data, timeperiod=21)

#         # Pandas_TA
#         self.qtpylib_ema = qtpylib.ema(
#             self.data.close,
#             window=21,
#             adjust=False,
#             min_periods=21 - 1,
#         )
#         self.qtpylib_ema_raw = qtpylib.ema(
#             self.raw_data.close,
#             window=21,
#             adjust=False,
#             min_periods=21 - 1,
#         )

#     def test_types(self):
#         self.assertIsInstance(
#             self.ta_ema,
#             pd.Series,
#             "TaLib EMA is not instance of pandas.Series",
#         )
#         self.assertIsInstance(
#             self.qtpylib_ema,
#             pd.Series,
#             "Pandas_ta EMA is not instance of pandas.Series",
#         )

#     def test_indicator(self):
#         pdtest.assert_series_equal(
#             self.ta_ema,
#             self.qtpylib_ema,
#             check_names=False,
#         )
#         pdtest.assert_series_equal(
#             self.ta_ema,
#             self.ta_ema_raw,
#             check_names=False,
#             check_index_type=False,
#         )
#         pdtest.assert_series_equal(
#             self.qtpylib_ema,
#             self.qtpylib_ema_raw,
#             check_names=False,
#             check_index_type=False,
#         )
#         pdtest.assert_series_equal(
#             self.ta_ema_raw,
#             self.qtpylib_ema_raw,
#             check_names=False,
#         )

#     def test_next(self):
#         df = self.data.copy(deep=True)
#         df["ema"] = self.ta_ema

#         for i in range(0, len(df)):
#             ema_value = df.l.ema[0]
#             raw_ema_value = self.ta_ema_raw.iloc[i]

#             # Next first to by pass next continue condiction
#             df.next()

#             if np.isnan(ema_value) and np.isnan(raw_ema_value):
#                 continue

#             self.assertEqual(ema_value, raw_ema_value, f"EMA[{i}] is wrong")


# if __name__ == "__main__":
#     unittest.main(verbosity=2)



from typing import Callable, Literal

import talib as ta
import talib.abstract as taa


def talib_ma_mode(
    name: Literal["sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3"]
) -> int:
    """Helper Function that returns the Enum value for TA Lib's MA Type"""
    if isinstance(name, str) and len(name) > 1:
        name = name.lower()
        if name == "sma":
            return ta.MA_Type.SMA  # 0
        elif name == "ema":
            return ta.MA_Type.EMA  # 1
        elif name == "wma":
            return ta.MA_Type.WMA  # 2
        elif name == "dema":
            return ta.MA_Type.DEMA  # 3
        elif name == "tema":
            return ta.MA_Type.TEMA  # 4
        elif name == "trima":
            return ta.MA_Type.TRIMA  # 5
        elif name == "kama":
            return ta.MA_Type.KAMA  # 6
        elif name == "mama":
            return ta.MA_Type.MAMA  # 7
        elif name == "t3":
            return ta.MA_Type.T3  # 8
    return 0  # Default: SMA -> 0


def talib_ma(
    name: Literal["sma", "ema", "wma", "dema", "tema", "trima", "kama", "mama", "t3"]
) -> Callable:
    """Helper Function that returns the Enum value for TA Lib's MA Type"""
    if isinstance(name, str) and len(name) > 1:
        name = name.lower()
        if name == "sma":
            return taa.SMA
        elif name == "ema":
            return taa.EMA
        elif name == "wma":
            return taa.WMA
        elif name == "dema":
            return taa.DEMA
        elif name == "tema":
            return taa.TEMA
        elif name == "trima":
            return taa.TRIMA
        elif name == "kama":
            return taa.KAMA
        elif name == "mama":
            return taa.MAMA
        elif name == "t3":
            return taa.T3
    return taa.SMA



import pandas as pd


def cdl_direction(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
) -> pd.Series | pd.DataFrame:
    """Direction of candle

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: `pd.Series` return [100, 0 , -100]
            - `100` for bull bar
            - `-100` for bear bar
            - `0` for None
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
    i = dataframe.apply(
        lambda r: 100 if r.open < r.close else -100 if r.open > r.close else 0,
        axis=1,
    ).astype(int)

    if inplace:
        dataframe[name or f"{prefix}direction"] = i
        return dataframe

    return i



import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from lettrade.data import DataFeed
from lettrade.plot import BotPlotter, PlotColor


class PlotlyBotPlotter(BotPlotter):
    """Class help to plot `lettrade`"""

    figure: go.Figure = None

    _data_shape: dict

    def stop(self):
        """stop plotter"""

    def _config_default(self):
        """Generate default config

        Example:
            ```json
            {
                "groups": [
                    {
                        "id": "EURUSD_5m",
                        "type": "data",
                        "data": pandas.DataFrame,
                        "height": 1,
                        "items": [
                            {
                                "type": "candlestick",
                                "show_orders": True,
                                "show_positions": True,
                                "update_xaxes": {
                                    "title_text": "EURUSD_5m",
                                    "rangeslider_visible": False,
                                    "mirror": True,
                                    "ticks": "outside",
                                    "showline": True
                                },
                                "update_yaxes": {
                                    "title_text": "Price $",
                                    "mirror": True,
                                    "ticks": "outside",
                                    "showline": True
                                }
                            },
                            {
                                "type": "scatter",
                                "x": pandas.DatetimeIndex,
                                "y": pandas.Series,
                                "line": { "color": <PlotColor.AMBER: '#fa0'>, "width": 1 },
                                "name": "ema1",
                                "mode": "lines"
                            },
                            {
                                "type": "scatter",
                                "x": pandas.DatetimeIndex,
                                "y": pandas.Series,
                                "line": { "color": <PlotColor.CYAN: '#00bad6'>, "width": 1 },
                                "name": "ema2",
                                "mode": "lines"
                            }
                        ]
                    },
                    { "id": "equity", "type": "equity", "height": 0.5 }
                ],
                "params": {
                    "shared_xaxes": True,
                    "vertical_spacing": 0.03,
                    "rows": 2,
                    "row_heights": [1, 0.5]
                },
                "layout": {
                    "xaxis": {
                        "rangeselector": {
                            "bgcolor": "#282a36",
                            "activecolor": "#5b5b66",
                            "buttons": [
                            { "step": "all" },
                            {
                                "count": 6,
                                "label": "6 day",
                                "step": "day",
                                "stepmode": "backward"
                            },
                            {
                                "count": 3,
                                "label": "3 day",
                                "step": "day",
                                "stepmode": "backward"
                            },
                            {
                                "count": 1,
                                "label": "1 day",
                                "step": "day",
                                "stepmode": "backward"
                            }
                            ]
                        }
                    },
                    "yaxis": { "autorange": True, "fixedrange": False },
                    "title": {
                        "text": "<__main__.SmaCross object at 0x709c53ee4d70>",
                        "font": { "size": 24 },
                        "x": 0.5,
                        "xref": "paper"
                    },
                    "modebar_add": [
                        "v1hovermode",
                        "hoverclosest",
                        "hovercompare",
                        "togglehover",
                        "togglespikelines",
                        "drawline",
                        "drawopenpath",
                        "drawclosedpath",
                        "drawcircle",
                        "drawrect",
                        "eraseshape"
                    ],
                    "hovermode": "x unified"
                },
                "shapes": {
                    "EURUSD_5m": { "group_index": 0, "rows": 1, "cols": 1 },
                    "equity": { "group_index": 1, "rows": 1, "cols": 1 }
                },
                "rows_total": 2,
                "cols_total": 2
            }
            ```

        Returns:
            _type_: _description_
        """
        # --- Data
        groups = []
        for data in self.datas:
            groups.append(
                dict(
                    id=data.name,
                    type="data",
                    data=data,
                    height=1,
                    items=[
                        dict(
                            type="candlestick",
                            name=data.name,
                            show_orders=data is self.data,
                            show_positions=data is self.data,
                            update_xaxes=dict(
                                title_text=data.name,
                                rangeslider_visible=False,
                                # showspikes=True,
                                # spikemode="across",
                                mirror=True,
                                ticks="outside",
                                showline=True,
                                # linecolor="",
                            ),
                            update_yaxes=dict(
                                title_text="Price $",
                                # autorange=True,
                                # fixedrange=False,
                                # showspikes=True,
                                # spikemode="across",
                                mirror=True,
                                ticks="outside",
                                showline=True,
                            ),
                        )
                    ],
                )
            )

        # --- Equity
        groups.append(
            dict(
                id="equity",
                type="equity",
                row_height=0.5,
            )
        )

        # --- Buttons
        buttons = [dict(step="all")]
        match self.data.timeframe.unit:
            case "m":
                count = 1
                step = "day"
            case "h":
                count = 7
                step = "day"
            case _:
                count = 1
                step = "month"
        buttons.extend(
            [
                dict(
                    count=6 * count,
                    label=f"{6*count} {step}",
                    step=step,
                    stepmode="backward",
                ),
                dict(
                    count=3 * count,
                    label=f"{3*count} {step}",
                    step=step,
                    stepmode="backward",
                ),
                dict(
                    count=count,
                    label=f"{count} {step}",
                    step=step,
                    stepmode="backward",
                ),
            ]
        )
        return dict(
            groups=groups,
            # --- Params
            params=dict(
                shared_xaxes=True,
                vertical_spacing=0.03,
            ),
            # --- Layout
            layout=dict(
                xaxis=dict(
                    rangeselector=dict(
                        bgcolor="#282a36",
                        activecolor="#5b5b66",
                        buttons=buttons,
                    ),
                ),
                yaxis=dict(
                    autorange=True,
                    fixedrange=False,
                ),
                title=dict(
                    text=str(self.strategy),
                    font=dict(size=24),
                    x=0.5,
                    xref="paper",
                ),
                # autosize=False,
                # width=800,
                height=1_000 * len(self.datas),
                modebar_add=[
                    "v1hovermode",
                    "hoverclosest",
                    "hovercompare",
                    "togglehover",
                    "togglespikelines",
                    "drawline",
                    "drawopenpath",
                    "drawclosedpath",
                    "drawcircle",
                    "drawrect",
                    "eraseshape",
                ],
                # template="plotly_dark",
                hovermode="x unified",
            ),
        )

    def _config_strategy(self, config: dict):
        """Merge strategy plot config"""
        strategy_config: dict = self.strategy._plot(*self.datas)

        # Move global items to main data items
        group_indexes = list(group["id"] for group in config["groups"])
        if "items" in strategy_config:
            group_index = group_indexes.index(self.data.name)
            data_group = config["groups"][group_index]
            data_items: list = data_group.setdefault("items", [])
            data_items.extend(strategy_config["items"])

        for data in self.datas:
            dname = data.name
            if dname not in strategy_config:
                continue

            group_index = group_indexes.index(data.name)
            data_group = config["groups"][group_index]
            data_config = strategy_config[dname]

            data_items: list = data_group.setdefault("items", [])
            data_items.extend(data_config["items"])

        if "layout" in strategy_config:
            config["layout"].update(strategy_config["layout"])

        return config

    def _config_standard(self, config: dict):
        """Calculate shapes/heights"""
        shapes = dict()
        rows_total = 1
        cols_total = 1
        row_heights = []
        for i, group in enumerate(config["groups"]):
            if group["type"] == "data":
                _rows = []
                _cols = []
                _heights = dict()
                for item in group["items"]:
                    row = item.get("row", 1)
                    _rows.append(row)
                    _cols.append(item.get("col", 1))
                    _heights[row] = item.pop("row_height", 1)

                rows = max(_rows)
                cols = max(_cols)

                for j in range(1, rows):
                    if j not in _heights.values():
                        _heights[j] = 1

                row_heights += _heights.values()
            else:
                rows = group.get("row", 1)
                cols = group.get("col", 1)
                row_heights += [group.pop("row_height", 1)]

            shapes[group["id"]] = dict(
                group_index=i,
                row=rows_total,
                col=cols_total,
                rows=rows,
                cols=cols,
            )
            rows_total += rows
            cols_total = cols

        params: dict = config.setdefault("params", dict())

        # rows -1: because row index start at 1
        params.update(rows=rows_total - 1, row_heights=row_heights)

        config.update(
            params=params,
            shapes=shapes,
            rows_total=rows_total,
            cols_total=cols_total,
        )
        return config

    def _plot_config(self, config: dict):
        """Apply config to figure"""
        rows = 0
        for group in config["groups"]:
            group_shape = config["shapes"][group["id"]]

            if group["type"] == "data":
                for item in group["items"]:
                    row = rows + item.pop("row", 1)
                    col = item.pop("col", 1)

                    match item["type"]:
                        case "candlestick":
                            self._plot_candlestick(
                                data=group["data"],
                                row=row,
                                col=col,
                                **item,
                            )
                        case "scatter":
                            self._plot_scatter(
                                data=group["data"],
                                row=row,
                                col=col,
                                **item,
                            )
            elif group["type"] == "equity":
                row = rows + group.pop("row", 1)
                col = group.pop("col", 1)

                self._plot_equity(row=row, col=col, **group)

            rows += group_shape["rows"]

    def _plot_candlestick(
        self,
        data: DataFeed,
        x=None,
        open=None,
        high=None,
        low=None,
        close=None,
        update_xaxes: dict | None = None,
        update_yaxes: dict | None = None,
        show_orders: bool = False,
        show_positions: bool = False,
        row: int = 1,
        col: int = 1,
        type: str | None = None,
        **kwargs,
    ):
        self.figure.add_trace(
            go.Candlestick(
                x=x if x is not None else data.index,
                open=open if open is not None else data["open"],
                high=high if high is not None else data["high"],
                low=low if low is not None else data["low"],
                close=close if close is not None else data["close"],
                **kwargs,
            ),
            row=row,
            col=col,
        )
        if update_xaxes is not None:
            self.figure.update_xaxes(row=row, col=col, **update_xaxes)
        if update_yaxes is not None:
            self.figure.update_yaxes(row=row, col=col, **update_yaxes)
        if show_orders:
            self._plot_orders(data=data, row=row, col=col)
        if show_positions:
            self._plot_positions(data=data, row=row, col=col)

    def _plot_scatter(
        self,
        data: DataFeed,
        x: pd.Index,
        y: pd.Series,
        row: int = 1,
        col: int = 1,
        fullfill: bool = False,
        type: str | None = None,
        **kwargs,
    ):
        # Fill None
        if fullfill and x is not data.index:
            s = pd.Series(None, index=data.index)
            s.loc[x] = y
            x = s.index
            y = s

        self.figure.add_scatter(x=x, y=y, row=row, col=col, **kwargs)

    def _plot_equity(
        self,
        row: int = 1,
        col: int = 1,
        type: str | None = None,
        **kwargs,
    ):
        equities = self.account._equities

        # Filter equities in data range only
        start_dt = self.data.index[0]
        stop_dt = self.data.index[-1]
        equities = {k: e for k, e in equities.items() if k > start_dt and k < stop_dt}

        if len(equities) == 0:
            return

        # Axis
        x = list(equities.keys())
        y = list(equities.values())

        self.figure.add_trace(
            go.Scatter(
                x=x,
                y=y,
                line=dict(color="#ff9900", width=2),
                # showlegend=False,
                name="Equity",
                # fill="toself",
            ),
            row=row,
            col=col,
        )
        self.figure.update_yaxes(
            title_text="Equity",
            row=row,
            col=col,
        )

    def _plot_orders(
        self,
        data: DataFeed,
        row: int = 1,
        col: int = 1,
    ):
        orders = list(self.exchange.history_orders.values()) + list(
            self.exchange.orders.values()
        )

        # Filter order in data range only
        start_dt = data.index[0]
        stop_dt = data.index[-1]
        orders = [o for o in orders if o.placed_at > start_dt and o.placed_at < stop_dt]

        # TODO: test using order/sl/tp line for performance
        for order in orders:
            x = [order.placed_at]
            y = [order.place_price]

            hovertemplate = (
                "<br>"
                # f"Order id: {order.id}<br>"
                "Index: %{x}<br>"
                f"At: {order.placed_at}<br>"
                "Price: %{y}<br>"
                f"Size: {order.size}<br>"
            )
            if order.sl:
                hovertemplate += f"SL: {order.sl}<br>"
            if order.tp:
                hovertemplate += f"TP: {order.tp}<br>"

            self.figure.add_scatter(
                x=x,
                y=y,
                hovertemplate=hovertemplate,
                mode="markers",
                name=f"Order-{order.id}",
                marker=dict(
                    symbol="line-ew-open",
                    size=10,
                    color=(PlotColor.GREEN if order.is_long else PlotColor.RED),
                ),
                showlegend=False,
                row=row,
                col=col,
            )

    def _plot_positions(
        self,
        data: DataFeed,
        row: int = 1,
        col: int = 1,
    ):
        positions = list(self.exchange.history_positions.values()) + list(
            self.exchange.positions.values()
        )

        # Filter equities in data range only
        start_dt = data.index[0]
        stop_dt = data.index[-1]
        _positions = []
        for p in positions:
            if p.entry_at < start_dt or p.entry_at > stop_dt:
                continue
            if p.exit_at:
                if p.exit_at < start_dt or p.exit_at > stop_dt:
                    continue
            _positions.append(p)
        positions = _positions

        for position in positions:
            x = [position.entry_at]
            y = [position.entry_price]
            customdata = [["Entry", position.size, position.pl, position.fee]]
            if position.exit_at:
                x.append(position.exit_at)
                y.append(position.exit_price)
                customdata.append(["Exit", position.size, position.pl, position.fee])

            color = PlotColor.CYAN if position.is_long else PlotColor.LIGHT_PINK
            self.figure.add_scatter(
                x=x,
                y=y,
                customdata=customdata,
                hovertemplate=(
                    "<br>"
                    "At: %{x}<br>"
                    "Price: %{y}<br>"
                    "Status: %{customdata[0]}<br>"
                    "Size: %{customdata[1]}<br>"
                    "PL: %{customdata[2]:.2f}$<br>"
                    "Fee: %{customdata[3]:.2f}$<br>"
                ),
                mode="lines+markers",
                name=f"Position-{position.id}",
                marker=dict(
                    symbol="circle-dot",
                    size=10,
                    color=color,
                ),
                line=dict(
                    color=color,
                    width=1,
                    dash="dash",
                ),
                showlegend=False,
                row=row,
                col=col,
            )

    def load(self):
        """Load plot config from `Strategy.plot()` and setup candlestick/equity"""

        config = self._config_default()
        config = self._config_strategy(config)
        config = self._config_standard(config)

        params: dict = config.setdefault("params", dict())
        self.figure = make_subplots(**params)

        self._plot_config(config)

        self.figure.update_layout(**config["layout"])

    def plot(self, jump: dict | None = None, **kwargs):
        """Plotly show figure

        Args:
            jump (dict | None, optional): _description_. Defaults to None.
        """
        if jump is not None:
            self.jump(**jump)
        elif self.figure is None:
            self.load()
        else:
            if self.jump_reset():
                self.load()

        params = dict(layout_xaxis_rangeslider_visible=False)
        params.update(**kwargs)
        self.figure.update(**params)

        self.figure.show()



from abc import ABCMeta, abstractmethod

from .data import DataFeed


class DataFeeder(metaclass=ABCMeta):
    datas: list[DataFeed]
    data: DataFeed

    def init(self, datas: list[DataFeed]):
        self.datas = datas
        self.data = datas[0]

    @property
    def is_continous(self):
        """Flag check is realtime continous datafeeder"""
        return True

    def alive(self):
        pass

    def start(self, **kwargs):
        """Load init datafeeds"""

    @abstractmethod
    def next(self):
        pass

    def stop(self):
        pass



import logging
import os
from collections.abc import Callable
from concurrent.futures import Future, ProcessPoolExecutor
from itertools import product, repeat
from multiprocessing import Queue
from typing import Any, Literal

import numpy as np
import pandas as pd

from lettrade import (
    Account,
    BotPlotter,
    BotStatistic,
    Commander,
    DataFeed,
    DataFeeder,
    Exchange,
    LetTrade,
    LetTradeBot,
    Strategy,
)

from .account import BackTestAccount
from .commander import BackTestCommander
from .data import BackTestDataFeed, CSVBackTestDataFeed
from .exchange import BackTestExchange
from .feeder import BackTestDataFeeder
from .plot import OptimizePlotter
from .stats import OptimizeStatistic

logger = logging.getLogger(__name__)


def logging_filter_optimize():
    # logging.getLogger("lettrade.exchange.backtest.backtest").setLevel(logging.WARNING)
    logging.getLogger("lettrade.exchange.backtest.commander").setLevel(logging.WARNING)
    logging.getLogger("lettrade.exchange.backtest.exchange").setLevel(logging.WARNING)
    logging.getLogger("lettrade.exchange.order").setLevel(logging.WARNING)
    logging.getLogger("lettrade.stats.stats").setLevel(logging.WARNING)
    logging.getLogger("lettrade.bot").setLevel(logging.WARNING)


class LetTradeBackTestBot(LetTradeBot):
    """LetTradeBot for backtest"""

    def init(self, optimize: dict[str, Any] = None, **kwargs):
        strategy_kwargs = self._kwargs.setdefault("strategy_kwargs", {})
        strategy_kwargs.update(is_optimize=optimize is not None)

        super().init()

        # Optimize
        if not optimize:
            return

        # Set optimize params
        for attr, value in optimize.items():
            setattr(self.strategy, attr, value)


class LetTradeBackTest(LetTrade):
    _stats: OptimizeStatistic = None

    @property
    def _optimize_stats_cls(self) -> type["OptimizeStatistic"]:
        return self._kwargs.get("optimize_stats_cls", None)

    @property
    def _optimize_plotter_cls(self) -> type["OptimizePlotter"]:
        return self._kwargs.get("optimize_plotter_cls", None)

    @_optimize_plotter_cls.setter
    def _optimize_plotter_cls(self, value):
        self._kwargs["optimize_plotter_cls"] = value

    @property
    def _strategy_cls(self) -> type[Strategy]:
        return self._kwargs.get("strategy_cls", None)

    def _datafeed(
        self,
        data: str | DataFeed | BackTestDataFeed | pd.DataFrame,
        index: int,
        **kwargs,
    ):
        match data:
            case str():
                data = CSVBackTestDataFeed(path=data)
            case BackTestDataFeed():
                pass
            case DataFeed():
                data = BackTestDataFeed(name=data.name, data=data)
            case pd.DataFrame():
                data = BackTestDataFeed(name=f"data_{index}", data=data)
            case _:
                raise RuntimeError(f"Data {data} type is invalid")

        return super()._datafeed(data, index=index, **kwargs)

    def start(self, force: bool = False):
        # Load plotly here just for backtest to improve performance
        if self._plotter_cls == "PlotlyBotPlotter":
            from lettrade.plot.plotly import PlotlyBotPlotter

            self._plotter_cls = PlotlyBotPlotter

        return super().start(force)

    def run(self, worker: int | None = None, **kwargs):
        # Load plotly here just for backtest to improve performance
        if self._plotter_cls == "PlotlyBotPlotter":
            from lettrade.plot.plotly import PlotlyBotPlotter

            self._plotter_cls = PlotlyBotPlotter

        return super().run(worker, **kwargs)

    def _optimize_init(self, cache: str, total: int, process_bar: bool):
        # Disable logging
        logging_filter_optimize()

        # Disable commander
        if self._commander_cls:
            self._commander_cls = None

        # Disable Bot Plotter
        if self._plotter_cls:
            self._plotter_cls = None

        # Enable Optimize plotter
        if self._optimize_plotter_cls == "PlotlyOptimizePlotter":
            from .plotly import PlotlyOptimizePlotter

            self._optimize_plotter_cls = PlotlyOptimizePlotter

        if self._optimize_plotter_cls is not None:
            self._plotter = self._optimize_plotter_cls(
                total=total,
                process_bar=process_bar,
                **self._kwargs.get("optimize_plotter_kwargs", {}),
            )

        # Enable Optimize stats
        self._stats = self._optimize_stats_cls(
            plotter=self._plotter,
            total=total,
            **self._kwargs.get("optimize_plotter_kwargs", {}),
        )

        # Optimize stats queue
        self._kwargs["queue"] = self._stats.queue

        # Optimize cache dir
        if cache is not None:
            cache = _optimize_cache_dir(cache, self._strategy_cls)
            self._kwargs["cache"] = cache

    # --- Optimize: Grid search
    def optimize(
        self,
        multiprocessing: Literal["auto", "fork"] = "auto",
        workers: int | None = None,
        process_bar: bool = True,
        cache: str = "data/optimize",
        **kwargs,
    ):
        """Backtest optimization

        Args:
            multiprocessing (str | None, optional): _description_. Defaults to "auto".
        """
        if self.data.l.pointer != 0:
            # TODO: Can drop unnecessary columns by snapshort data.columns from init time
            raise RuntimeError(
                "Optimize datas is not clean, don't run() backtest before optimize()"
            )

        # optimizes = list(product(*(zip(repeat(k), v) for k, v in kwargs.items())))
        optimizes = list(
            dict(zip(kwargs.keys(), values)) for values in product(*kwargs.values())
        )

        self._optimize_init(cache=cache, total=len(optimizes), process_bar=process_bar)

        # Run optimize in multiprocessing
        self._optimizes_multiproccess(
            optimizes=optimizes,
            multiprocessing=multiprocessing,
            workers=workers,
        )

        self.optimize_done()

    def _optimizes_multiproccess(
        self,
        optimizes: list[set],
        multiprocessing: Literal["auto", "fork"],
        workers: int | None = None,
    ):
        # If multiprocessing start method is 'fork' (i.e. on POSIX), use
        # a pool of processes to compute results in parallel.
        # Otherwise (i.e. on Windows), sequential computation will be "faster".
        if multiprocessing == "fork" or (
            multiprocessing == "auto" and os.name == "posix"
        ):
            optimizes_batches = list(_batch(optimizes, workers=workers))

            # Set max workers
            if workers is None or workers > len(optimizes_batches):
                workers = len(optimizes_batches)
                logger.info("Set optimize workers to %d", workers)

            with ProcessPoolExecutor(max_workers=workers) as executor:
                futures: list[Future] = []
                index = 0
                for optimizes in optimizes_batches:
                    future = executor.submit(
                        self.__class__._optimizes_run,
                        optimizes=optimizes,
                        index=index,
                        **self._kwargs,
                    )
                    futures.append(future)
                    index += len(optimizes)

                for future in futures:
                    future.result()
        else:
            if os.name == "posix":
                logger.warning(
                    "For multiprocessing support in `optimize()` "
                    "set multiprocessing='fork'."
                )

            self.__class__._optimizes_run(
                optimizes=optimizes,
                **self._kwargs,
            )

    @classmethod
    def _optimizes_run(
        cls,
        datas: list[DataFeed],
        optimizes: list[dict[str, Any]],
        index: int = 0,
        **kwargs,
    ):
        """Run optimize in class method to not copy whole LetTradeBackTest self object

        Args:
            datas (list[DataFeed]): _description_
            optimizes (list[dict]): _description_
            index (int, optional): _description_. Defaults to 0.

        Returns:
            _type_: _description_
        """
        results = []
        for i, optimize in enumerate(optimizes):
            result = cls._optimize_run(
                datas=[d.copy(deep=True) for d in datas],
                optimize=optimize,
                index=index + i,
                **kwargs,
            )
            if result is not None:
                results.append(result)
        return results

    # --- Optimize: model for external optimizer
    # Create optimize model environment
    _opt_main_pid: int = None
    _opt_params_parser: Callable[[Any], list[set[str, Any]]] = None
    _opt_result_parser: Callable[[pd.Series], float] = None
    _opt_kwargs: dict = None

    def optimize_model(
        self,
        params_parser: Callable[[Any], list[set[str, Any]]] = None,
        result_parser: Callable[[pd.Series], float] = None,
        total: int = 0,
        cache: str = "data/optimize",
        process_bar: bool = False,
        dumper: Callable[[dict, "LetTradeBackTest"], None] | None = None,
    ) -> Callable[[Any], Any]:
        """Optimize function help to integrated with external optimize trainer

        Args:
            params_parser (Callable[[Any], list[set[str, Any]]], optional): Parse external parameters to bot parameters dict. Defaults to None.
            result_parser (Callable[[pd.Series], float], optional): Parse bot result to external score. Defaults to None.
            total (int, optional): Total number of optimize if possible. Defaults to 0.
            cache (str, optional): Cache directory to store optimize result. Defaults to "data/optimize".
            process_bar (bool, optional): Enable/Disable process bar. Defaults to False.

        Raises:
            RuntimeError: _description_

        Returns:
            Callable[[Any], Any]: Optimize model function
        """
        if self.data.l.pointer != 0:
            raise RuntimeError(
                "Optimize datas is not clean, don't run() backtest before optimize()"
            )

        self._optimize_init(cache=cache, total=total, process_bar=process_bar)

        # Optimize parameters
        optimizer_kwargs = dict(
            main_pid=os.getpid(),
            params_parser=params_parser,
            result_parser=result_parser,
            kwargs=self._kwargs,
        )

        if dumper is not None:
            dumper(optimizer_kwargs, self)
        else:
            self.__class__._optimize_model_kwargs(optimizer_kwargs)

        return self.__class__._optimize_model

    @classmethod
    def _optimize_model_kwargs(cls, kwargs: dict):
        for key, value in kwargs.items():
            setattr(cls, f"_opt_{key}", value)

    @classmethod
    def _optimize_model(cls, optimize: dict[str, Any], optimizer_kwargs=None, **kwargs):
        """Model to run bot in singleprocessing or multiprocessing

        Args:
            optimize (list[set[str, Any]]): _description_

        Raises:
            RuntimeError: _description_

        Returns:
            _type_: _description_
        """
        # Optimizer loader
        if optimizer_kwargs is not None:
            cls._optimize_model_kwargs(optimizer_kwargs)

        # Load optimize parameters
        if cls._opt_params_parser:
            optimize = cls._opt_params_parser(optimize)

        # Load cache here to skip copy kwargs and data data
        cache = cls._opt_kwargs.get("cache", None)
        if cache is not None:
            cached = _optimize_cache_get(dir=cache, optimize=optimize)
            if cached is not None:
                result = cached["result"]

                # Put result to stats
                queue = cls._opt_kwargs.get("queue", None)
                if queue is not None:
                    queue.put(dict(index=None, optimize=optimize, result=result))

                # Prepare model result
                if cls._opt_result_parser:
                    result = cls._opt_result_parser(result)

                logger.info("Optimize load cache: %s", cached["path"])
                return result

        # If models run in singleprocessing, copy kwargs for bot to not overrite main kwargs
        if os.getpid() == cls._opt_main_pid:
            opt_kwargs = cls._opt_kwargs.copy()
        else:
            opt_kwargs = cls._opt_kwargs

        # Check data didn't reload by multiprocessing
        datas = opt_kwargs.pop("datas")
        data = datas[0]
        if data.l.pointer != 0:
            print(data.l.pointer, data.l)
            raise RuntimeError(
                "Optimize model data changed, set fork_data=True to reload"
            )

        datas = [d.copy(deep=True) for d in datas]

        # Run
        result = cls._optimize_run(
            datas=datas,
            optimize=optimize,
            **opt_kwargs,
        )
        if cls._opt_result_parser:
            result = cls._opt_result_parser(result)

        return result

    # --- Optimize: load cached optimize result
    def optimize_cache(self, cache: str = "data/optimize"):
        """Load optimize results from cache

        Args:
            cache (str, optional): Cache directory. Defaults to "data/optimize".
        """
        import json

        self._optimize_init(cache=cache, total=0, process_bar=False)
        cache_dir = self._kwargs["cache"]
        queue = self._kwargs["queue"]

        logger.info("Load caches from: %s", cache)

        for cache_file in os.listdir(cache_dir):
            if cache_file == "info.json":
                continue

            try:
                cache_path = f"{cache_dir}/{cache_file}"
                data = json.load(open(cache_path, encoding="utf-8"))

                queue.put(
                    dict(
                        index=cache_file,
                        optimize=data["optimize"],
                        result=data["result"],
                    )
                )
            except Exception as e:
                logger.warning("Loading cache %s error %s", cache_path, e)

        logger.info("Loaded %s caches", len(self._stats.results))

    def optimize_done(self):
        """Clean and close optimize handlers"""
        self._stats.done()

    # --- Optimize: run
    @classmethod
    def _optimize_run(
        cls,
        datas: list[DataFeed],
        optimize: dict[str, Any],
        bot_cls: type[LetTradeBot],
        index: int = 0,
        queue: "Queue | None" = None,
        cache: str = None,
        **kwargs,
    ):
        try:
            # Load cache
            # TODO: load cache at beginning of caller
            if cache is not None:
                cached = _optimize_cache_get(dir=cache, optimize=optimize)
                if cached is not None:
                    result = cached["result"]

                    # Put result to stats
                    if queue is not None:
                        queue.put(dict(index=index, optimize=optimize, result=result))

                    # logger.info("Optimize load cache: %s", cached["path"])
                    return result

            # Load bot
            if datas:
                kwargs["datas"] = datas

            bot = bot_cls.run_bot(
                optimize=optimize,
                id=index,
                init_kwargs=dict(optimize=optimize),
                result="bot",
                **kwargs,
            )
            result = bot.stats.result

            if cache is not None:
                _optimize_cache_set(dir=cache, optimize=optimize, result=result)

            if queue is not None:
                queue.put(dict(index=index, optimize=optimize, result=result))

            return result
        except Exception as e:
            logger.error("Optimize %d", index, exc_info=e)
            raise e


def _batch(seq, workers=None):
    n = np.clip(int(len(seq) // (workers or os.cpu_count() or 1)), 1, 300)
    for i in range(0, len(seq), n):
        yield seq[i : i + n]


def _md5_dict(d: dict):
    import hashlib
    import json

    return hashlib.md5(json.dumps(d, sort_keys=True).encode("utf-8")).hexdigest()


def _optimize_cache_dir(dir: str, strategy_cls: type[Strategy]) -> str:
    import hashlib
    import inspect
    import json
    from importlib.metadata import version
    from pathlib import Path

    info = dict(
        lettrade=version("lettrade"),
        strategy=str(strategy_cls),
    )

    try:
        strategy_file = inspect.getfile(strategy_cls)
        info.update(
            strategy_file=strategy_file,
            strategy_hash=hashlib.md5(open(strategy_file, "rb").read()).hexdigest(),
        )
    except OSError:
        # Get disassembly code of class
        import dis
        import io
        from contextlib import redirect_stdout

        with redirect_stdout(io.StringIO()) as f:
            dis.dis(strategy_cls)
            asm = f.getvalue()

        info.update(
            strategy_hash=hashlib.md5(asm.encode("utf-8")).hexdigest(),
        )

    cache_dir = Path(f"{dir}/{_md5_dict(info)}")
    cache_dir.mkdir(parents=True, exist_ok=True)

    # Strategy information
    with open(f"{cache_dir}/info.json", "w", encoding="utf-8") as f:
        json.dump(info, f)

    logger.info("Optimize cache directory: %s %s", cache_dir, info)
    return cache_dir.absolute()


def _optimize_cache_get(dir: str, optimize: dict) -> str | None:
    import json

    try:
        hash = _md5_dict(optimize)
        path = f"{dir}/{hash}.json"

        data = json.load(open(path, encoding="utf-8"))
        data["path"] = path
        data["result"] = pd.Series(data["result"])
        return data
    except FileNotFoundError:
        return None


def _optimize_cache_set(dir: str, optimize: dict, result: pd.Series):
    import json

    hash = _md5_dict(optimize)
    path = f"{dir}/{hash}.json"
    with open(path, "w+", encoding="utf-8") as f:
        json.dump(
            dict(
                optimize=optimize,
                result=json.loads(result.to_json()),
            ),
            f,
        )


def let_backtest(
    datas: DataFeed | list[DataFeed] | str | list[str],
    strategy: type[Strategy],
    *,
    feeder: type[DataFeeder] = BackTestDataFeeder,
    exchange: type[Exchange] = BackTestExchange,
    account: type[Account] = BackTestAccount,
    commander: type[Commander] | Commander | None = BackTestCommander,
    stats: type[BotStatistic] | None = BotStatistic,
    optimize_stats: type[OptimizeStatistic] | None = OptimizeStatistic,
    plotter: type[BotPlotter] | None = "PlotlyBotPlotter",
    optimize_plotter: type[OptimizePlotter] | None = "PlotlyOptimizePlotter",
    bot: type[LetTradeBackTestBot] | None = LetTradeBackTestBot,
    name: str | None = None,
    # Account kwargs
    balance: float | None = 10_000,
    commission: float | None = 0.2,
    leverage: float | None = 20,
    **kwargs,
) -> "LetTradeBackTest":
    """Complete `lettrade` backtest depenencies

    Args:
        datas (DataFeed | list[DataFeed] | str | list[str]): _description_
        strategy (type[Strategy]): The Strategy implement class
        feeder (type[DataFeeder], optional): _description_. Defaults to BackTestDataFeeder.
        exchange (type[Exchange], optional): _description_. Defaults to BackTestExchange.
        account (type[Account], optional): _description_. Defaults to BackTestAccount.
        commander (type[Commander] | Commander | None, optional): _description_. Defaults to BackTestCommander.
        stats (type[BotStatistic] | None, optional): _description_. Defaults to BotStatistic.
        optimize_stats (type[OptimizeStatistic] | None, optional): _description_. Defaults to OptimizeStatistic.
        plotter (type[BotPlotter] | None, optional): _description_. Defaults to "PlotlyBotPlotter".
        optimize_plotter (type[OptimizePlotter] | None, optional): _description_. Defaults to "PlotlyOptimizePlotter".
        bot (type[LetTradeBackTestBot] | None, optional): _description_. Defaults to LetTradeBackTestBot.
        name (str | None, optional): _description_. Defaults to None.
        balance (float | None, optional): _description_. Defaults to 10_000.
        commission (float | None, optional): _description_. Defaults to 0.2.
        leverage (float | None, optional): _description_. Defaults to 20.

    Returns:
        LetTradeBackTest: The LetTrade backtesting object
    """

    account_kwargs: dict = kwargs.setdefault("account_kwargs", {})
    account_kwargs.update(
        balance=balance,
        commission=commission,
        leverage=leverage,
    )

    return LetTradeBackTest(
        strategy=strategy,
        datas=datas,
        feeder=feeder,
        exchange=exchange,
        commander=commander,
        account=account,
        stats=stats,
        plotter=plotter,
        bot=bot,
        name=name,
        # Backtest
        optimize_stats_cls=optimize_stats,
        optimize_plotter_cls=optimize_plotter,
        **kwargs,
    )



import logging
import re

import pandas as pd

from lettrade.data import DataFeed

logger = logging.getLogger(__name__)


class BackTestDataFeed(DataFeed):
    """BackTest DataFeed"""

    def __init__(
        self,
        data: pd.DataFrame,
        name: str,
        timeframe: str | int | pd.Timedelta | None = None,
        meta: dict | None = None,
        since: int | str | pd.Timestamp | None = None,
        to: int | str | pd.Timestamp | None = None,
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            data (pd.DataFrame): _description_
            name (str): _description_
            timeframe (str | int | pd.Timedelta | None, optional): _description_. Defaults to None.
            meta (dict | None, optional): _description_. Defaults to None.
            since (int | str | pd.Timestamp | None, optional): Drop data before since. Defaults to None.
            to (int | str | pd.Timestamp | None, optional): Drop data after to. Defaults to None.
        """
        if timeframe is None:
            timeframe = self._find_timeframe(data)
            logger.info("DataFeed %s auto detect timeframe %s", name, timeframe)
        super().__init__(
            data=data,
            name=name,
            timeframe=timeframe,
            meta=meta,
            **kwargs,
        )
        if since is not None or to is not None:
            self.drop(since=since, to=to)

    def _find_timeframe(self, df):
        if len(df.index) < 3:
            raise RuntimeError("DataFeed not enough data to detect timeframe")
        dtl = df.index
        for i in range(0, 3):
            tf = dtl[i + 1] - dtl[i]
            if tf == dtl[i + 2] - dtl[i + 1]:
                return tf
        raise RuntimeError("DataFeed cannot detect timeframe")

    def alive(self):
        return self.l.pointer_stop > 1

    def next(
        self,
        size: int = 1,
        next: pd.Timestamp | None = None,
        missing="bypass",
    ) -> bool:
        has_next = True
        if not self.is_main:
            if next is None:
                raise RuntimeError("DataFeed parameter next is None")

            size = 0
            while True:
                try:
                    dt_next = self.l.index[size + 1]
                    # No more next data
                    # if not dt_next:
                    #     has_next = False
                    #     break

                    if dt_next > next:
                        break
                    size += 1
                except (KeyError, IndexError):
                    has_next = False
                    break

            # validate
            now = self.l.index[size]
            floor = self.timeframe.floor(next)
            if now != floor and missing != "bypass":
                raise RuntimeError(
                    f"DataFeed {self.name}: jump from [{now} to {self.l.index[size+1]}]"
                    f" missing range [{floor} to {next}]",
                )

        if size > 0:
            super().next(size=size)
        return has_next


_data_name_pattern = re.compile(r"^([\w\_]+)")


def _path_to_name(path: str):
    if "/" in path:
        path = path.rsplit("/", 1)[1]
    if "." in path:
        path = path.split(".", 1)[0]

    searchs = _data_name_pattern.search(path)
    if searchs:
        path = searchs.group(1)

    return path


class CSVBackTestDataFeed(BackTestDataFeed):
    """Implement help to load DataFeed from csv file"""

    def __init__(
        self,
        path: str | None = None,
        csv: dict | None = None,
        name: str | None = None,
        timeframe: str | int | pd.Timedelta | None = None,
        meta: dict | None = None,
        data: DataFeed | None = None,
        **kwargs: dict,
    ) -> None:
        """_summary_

        Args:
            path (str | None, optional): Path to csv file. Defaults to None.
            csv (dict | None, optional): Reflect of `pandas.read_csv()` parameters. Defaults to None.
            name (str | None, optional): _description_. Defaults to None.
            timeframe (str | int | pd.Timedelta | None, optional): _description_. Defaults to None.
            meta (dict | None, optional): _description_. Defaults to None.
            data (DataFeed | None, optional): _description_. Defaults to None.
            **kwargs (dict): [DataFeed](../../data/data.md#lettrade.data.data.DataFeed) dict parameters
        """
        if name is None:
            name = _path_to_name(path)

        if data is None:
            csv_params = dict(
                index_col=0,
                parse_dates=["datetime"],
                delimiter=",",
                header=0,
            )
            if csv is not None:
                csv_params.update(**csv)

            data = pd.read_csv(path, **csv_params)

            if not isinstance(data.index, pd.DatetimeIndex):
                data.index = data.index.astype("datetime64[ns, UTC]")

        super().__init__(
            data=data,
            name=name,
            timeframe=timeframe,
            meta=meta,
            **kwargs,
        )


class YFBackTestDataFeed(BackTestDataFeed):
    """YahooFinance DataFeed"""

    def __init__(
        self,
        name: str,
        ticker: str,
        since: str,
        to: str | None = None,
        period: str | None = None,
        interval: str = "1d",
        **kwargs,
    ) -> None:
        """_summary_

        Args:
            name (str): _description_
            ticker (str): _description_
            since (str): _description_
            to (str | None, optional): _description_. Defaults to None.
            period (str | None, optional): _description_. Defaults to None.
            interval (str, optional): _description_. Defaults to "1d".
        """
        params = dict(
            start=since,
            end=to,
            period=period,
            interval=interval,
        )

        # Download
        import yfinance as yf

        df = yf.download(ticker, **params)

        # Parse to lettrade datafeed
        from .extra.yfinance import yf_parse

        df = yf_parse(df)

        # Metadata
        meta = dict(yf=dict(ticker=ticker, **params))

        super().__init__(
            name=name,
            meta=meta,
            data=df,
            **kwargs,
        )



import unittest

import pandas as pd

from lettrade.data import TimeFrame


class DataFeedInjectTestCase(unittest.TestCase):
    def test_int(self):
        tf = TimeFrame(1)
        self.assertEqual(tf.string, "1m")

    def test_string(self):
        tf = TimeFrame("3s")
        self.assertEqual(tf.string, "3s")

        tf = TimeFrame("5m")
        self.assertEqual(tf.string, "5m")

        tf = TimeFrame("2h")
        self.assertEqual(tf.string, "2h")

        tf = TimeFrame("12h")
        self.assertEqual(tf.string, "12h")

        tf = TimeFrame("1d")
        self.assertEqual(tf.string, "1d")

        tf = TimeFrame("3d")
        self.assertEqual(tf.string, "3d")

    def test_timedelta(self):
        tf = TimeFrame(pd.Timedelta(minutes=3))
        self.assertEqual(tf.string, "3m")

        tf = TimeFrame(pd.Timedelta(hours=3))
        self.assertEqual(tf.string, "3h")

        tf = TimeFrame(pd.Timedelta(days=1))
        self.assertEqual(tf.string, "1d")

        tf = TimeFrame(pd.Timedelta(weeks=1))
        self.assertEqual(tf.string, "1w")

    def test_list(self):
        tf = TimeFrame([3, "s"])
        self.assertEqual(tf.string, "3s")

        tf = TimeFrame([1, "m"])
        self.assertEqual(tf.string, "1m")

        tf = TimeFrame([7, "h"])
        self.assertEqual(tf.string, "7h")

        tf = TimeFrame([1, "w"])
        self.assertEqual(tf.string, "1w")

    def test_exception(self):
        with self.assertRaises(RuntimeError):
            TimeFrame("asd123")

        with self.assertRaises(RuntimeError):
            TimeFrame("1a")

        with self.assertRaises(RuntimeError):
            TimeFrame("1z")

        with self.assertRaises(RuntimeError):
            TimeFrame(pd.Timestamp("2020-01-01"))

        with self.assertRaises(RuntimeError):
            TimeFrame("12d")

        with self.assertRaises(RuntimeError):
            TimeFrame("-1d")

    def test_floor(self):
        # Second
        tf = TimeFrame("5s")

        floor = tf.floor(pd.Timestamp("2020-01-01 01:00:03"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 01:00:00"))

        floor = tf.floor(pd.Timestamp("2020-01-01 01:00:23"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 01:00:20"))

        # Minute
        tf = TimeFrame("5m")

        floor = tf.floor(pd.Timestamp("2020-01-01 01:00:03"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 01:00:00"))

        floor = tf.floor(pd.Timestamp("2020-01-01 01:12:23"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 01:10:00"))

        # Hour
        tf = TimeFrame("4h")

        floor = tf.floor(pd.Timestamp("2020-01-01 11:11:03"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 8:00:00"))

        floor = tf.floor(pd.Timestamp("2020-01-01 21:12:23"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 20:00:00"))

        # Hour
        tf = TimeFrame("3h")

        floor = tf.floor(pd.Timestamp("2020-01-01 02:11:03"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 00:00:00"))

        floor = tf.floor(pd.Timestamp("2020-01-01 21:12:23"))
        self.assertEqual(floor, pd.Timestamp("2020-01-01 21:00:00"))

        # # Hour
        # tf = TimeFrame("7h")

        # floor = tf.floor(pd.Timestamp("2020-01-01 02:11:03"))
        # self.assertEqual(floor, pd.Timestamp("2020-01-01 00:00:00"))

        # floor = tf.floor(pd.Timestamp("2020-01-01 15:12:23"))
        # self.assertEqual(floor, pd.Timestamp("2020-01-01 14:00:00"))

        # Week
        tf = TimeFrame("1w")

        floor = tf.floor(pd.Timestamp("2020-01-01 11:11:03"))
        self.assertEqual(floor, pd.Timestamp("2019-12-30 00:00:00"))

        floor = tf.floor(pd.Timestamp("2020-03-03 21:12:23"))
        self.assertEqual(floor, pd.Timestamp("2020-03-02 00:00:00"))

    def test_ceil(self):
        # Second
        tf = TimeFrame("5s")

        ceil = tf.ceil(pd.Timestamp("2020-01-01 01:00:03"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-01 01:00:05"))

        ceil = tf.ceil(pd.Timestamp("2020-01-01 01:00:23"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-01 01:00:25"))

        # Minute
        tf = TimeFrame("5m")

        ceil = tf.ceil(pd.Timestamp("2020-01-01 01:00:03"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-01 01:05:00"))

        ceil = tf.ceil(pd.Timestamp("2020-01-01 01:12:23"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-01 01:15:00"))

        # Hour
        tf = TimeFrame("4h")

        ceil = tf.ceil(pd.Timestamp("2020-01-01 11:11:03"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-01 12:00:00"))

        ceil = tf.ceil(pd.Timestamp("2020-01-01 21:12:23"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-02 00:00:00"))

        # Day
        tf = TimeFrame("1d")

        ceil = tf.ceil(pd.Timestamp("2020-01-01 11:11:03"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-02 00:00:00"))

        ceil = tf.ceil(pd.Timestamp("2020-03-03 21:12:23"))
        self.assertEqual(ceil, pd.Timestamp("2020-03-04 00:00:00"))

        # Week
        tf = TimeFrame("1w")

        ceil = tf.ceil(pd.Timestamp("2020-01-01 11:11:03"))
        self.assertEqual(ceil, pd.Timestamp("2020-01-06 00:00:00"))

        ceil = tf.ceil(pd.Timestamp("2020-03-03 21:12:23"))
        self.assertEqual(ceil, pd.Timestamp("2020-03-09 00:00:00"))


if __name__ == "__main__":
    unittest.main(verbosity=2)



import logging

from lettrade.commander import Commander

logger = logging.getLogger(__name__)


class BackTestCommander(Commander):
    """
    BackTest Commander to debug commander notify
    """

    def start(self):
        logger.debug("BackTestCommander start")

    def stop(self):
        logger.debug("BackTestCommander stop")

    def send_message(self, msg: str, **kwargs):
        logger.info("BackTestCommander send_message: %s, %s", msg, kwargs)


class StorageBackTestCommander(Commander):
    """
    BackTest Commander to debug commander notify
    """

    storage = []

    def start(self):
        logger.debug("StorageBackTestCommander start")

    def stop(self):
        logger.debug("StorageBackTestCommander stop")

    def send_message(self, msg: str, **kwargs):
        logger.info("StorageBackTestCommander send_message: %s, %s", msg, kwargs)
        self.storage.append({"message": msg, **kwargs})



from importlib.util import find_spec

if find_spec("ray") is not None:
    from ray.util.queue import Queue

    class LetOptimizeRay:
        _data: dict

        def dumps(self, data: dict, lt: "LetTradeBackTest"):
            self._data = data

            # Patch: overrite lettrade queue
            q = Queue()
            data["kwargs"]["queue"] = q
            lt._stats._q = q

        @property
        def data(self) -> dict:
            return self._data



from typing import Literal

import pandas as pd
import talib.abstract as ta


def _plot_pattern(
    dataframe: pd.DataFrame,
    indicator: pd.Series,
    name: str | None = None,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
):
    if plot_kwargs is None:
        plot_kwargs = dict()

    plot_kwargs.update(name=name)

    from lettrade.indicator.plot import IndicatorPlotter
    from lettrade.plot.plotly import plot_candlestick, plot_mark

    filter = lambda df: df[name] > 0

    if plot_type == "mark":
        plot_kwargs.update(series=indicator)
        IndicatorPlotter(
            dataframe=dataframe,
            plotter=plot_mark,
            filter=filter,
            **plot_kwargs,
        )
    else:
        IndicatorPlotter(
            dataframe=dataframe,
            plotter=plot_candlestick,
            filter=filter,
            **plot_kwargs,
        )


def cdl_pattern(
    dataframe: pd.DataFrame,
    pattern: str,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): pandas.DataFrame with ohlcv
        pattern (str): TA-Lib candle pattern name. Ex: `3whitesoldiers`, `3blackcrows`
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )
        if plot and not inplace:
            raise RuntimeError("Cannot plot when inplace=False")

    i = getattr(ta, f"CDL{pattern.upper()}")(dataframe, **kwargs)

    if inplace:
        name = name or f"{prefix}{pattern.lower()}"
        dataframe[name] = i

        # Plot
        if plot:
            _plot_pattern(
                dataframe=dataframe,
                indicator=i,
                name=name,
                plot_type=plot_type,
                plot_kwargs=plot_kwargs,
            )

        return dataframe

    return i


def cdl_3blackcrows(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="3blackcrows",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_3whitesoldiers(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="3whitesoldiers",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_eveningstar(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="eveningstar",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_morningstar(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="morningstar",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_doji(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="doji",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_3inside(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="3inside",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )


def cdl_engulfing(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    plot: bool | list[str] = False,
    plot_type: Literal["candlestick", "mark"] = "candlestick",
    plot_kwargs: dict | None = None,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="engulfing",
        name=name,
        prefix=prefix,
        inplace=inplace,
        plot=plot,
        plot_type=plot_type,
        plot_kwargs=plot_kwargs,
        **kwargs,
    )



from .qtpylib import *


def inject_indicators():
    from pandas.core.base import PandasObject

    PandasObject.session = session
    PandasObject.atr = atr
    PandasObject.bollinger_bands = bollinger_bands
    PandasObject.cci = cci
    PandasObject.crossed = crossed
    PandasObject.crossed_above = crossed_above
    PandasObject.crossed_below = crossed_below
    PandasObject.heikinashi = heikinashi
    PandasObject.hull_moving_average = hull_moving_average
    PandasObject.ibs = ibs
    PandasObject.implied_volatility = implied_volatility
    PandasObject.keltner_channel = keltner_channel
    PandasObject.log_returns = log_returns
    PandasObject.macd = macd
    PandasObject.returns = returns
    PandasObject.roc = roc
    PandasObject.rolling_max = rolling_max
    PandasObject.rolling_min = rolling_min
    PandasObject.rolling_mean = rolling_mean
    PandasObject.rolling_std = rolling_std
    PandasObject.rsi = rsi
    PandasObject.stoch = stoch
    PandasObject.zscore = zscore
    PandasObject.pvt = pvt
    PandasObject.chopiness = chopiness
    PandasObject.tdi = tdi
    PandasObject.true_range = true_range
    PandasObject.mid_price = mid_price
    PandasObject.typical_price = typical_price
    PandasObject.vwap = vwap
    PandasObject.rolling_vwap = rolling_vwap
    PandasObject.weighted_bollinger_bands = weighted_bollinger_bands
    PandasObject.rolling_weighted_mean = rolling_weighted_mean

    PandasObject.sma = sma
    PandasObject.wma = wma
    PandasObject.ema = wma
    PandasObject.hma = hma

    PandasObject.zlsma = zlsma
    PandasObject.zlwma = zlema
    PandasObject.zlema = zlema
    PandasObject.zlhma = zlhma
    PandasObject.zlma = zlma



from lettrade.plot.plot import Plotter


class OptimizePlotter(Plotter):
    """
    Class help to plot `lettrade`
    """

    results: list

    def __init__(self) -> None:
        super().__init__()

    def init(self, results: list):
        self.results = results

    def on_result(self, result):
        """"""

    def on_done(self):
        """"""

    def heatmap(self, x: str, y: str, z: str = "equity", **kwargs):
        """"""

    def contour(self, x: str, y: str, z: str = "equity", **kwargs):
        """"""



import logging
from typing import TYPE_CHECKING

import pandas as pd

from .base import BaseTransaction, OrderState, OrderType
from .error import LetOrderValidateException

if TYPE_CHECKING:
    from lettrade import DataFeed, Exchange, Position

logger = logging.getLogger(__name__)


class Order(BaseTransaction):
    def __init__(
        self,
        id: str,
        exchange: "Exchange",
        data: "DataFeed",
        size: float,
        state: OrderState = OrderState.Pending,
        type: OrderType = OrderType.Market,
        limit_price: float | None = None,
        stop_price: float | None = None,
        sl_price: float | None = None,
        tp_price: float | None = None,
        parent: "Position | None" = None,
        tag: str | None = None,
        placed_at: pd.Timestamp | None = None,
        **kwargs,
    ):
        super().__init__(
            id=id,
            exchange=exchange,
            data=data,
            size=size,
            **kwargs,
        )

        self.type: OrderType = type
        self.state: OrderState = state
        self.limit_price: float | None = limit_price
        self.stop_price: float | None = stop_price
        self.sl_price: float | None = sl_price
        self.tp_price: float | None = tp_price
        self.parent: "Position | None" = parent
        self.tag: str | None = tag
        self.placed_at: pd.Timestamp | None = placed_at
        self.filled_at: pd.Timestamp | None = None
        self.filled_price: float | None = None

    def _repr_params(self):
        params = (
            f"id='{self.id}'"
            f", type='{self.type}'"
            f", state='{self.state}'"
            f", placed_at={self.placed_at}"
            # f", place_price={round(self.place_price, 5)}"
            f", size={round(self.size, 5)}"
        )
        if self.limit:
            params += f", limit={round(self.limit, 5)}"
        if self.stop:
            params += f", stop={round(self.stop, 5)}"
        if self.sl:
            params += f", sl={round(self.sl, 5)}"
        if self.tp:
            params += f", tp={round(self.tp, 5)}"

        if self.filled_at:
            params += (
                f", filled_at={self.filled_at}" f", filled_price={self.filled_price}"
            )
        if self.tag:
            params += f", tag='{self.tag}'"

        return params

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._repr_params()}>"

    def validate(self):
        # Validate
        price = self.place_price or self.data.l.open[0]

        # Buy side
        if self.size > 0:
            if self.sl_price is not None:
                if self.sl_price >= price:
                    raise LetOrderValidateException(
                        f"Order {self.id} BUY sl {self.sl_price} >= price {price}"
                    )
            if self.tp_price is not None:
                if self.tp_price <= price:
                    raise LetOrderValidateException(
                        f"Order {self.id} BUY tp {self.tp_price} <= price {price}"
                    )
        # Sell side
        elif self.size < 0:
            if self.sl_price is not None:
                if self.sl_price <= price:
                    raise LetOrderValidateException(
                        f"Order {self.id} SELL sl {self.sl_price} <= price {price}"
                    )
            if self.tp_price is not None:
                if self.tp_price >= price:
                    raise LetOrderValidateException(
                        f"Order {self.id} SELL tp {self.tp_price} >= price {price}"
                    )
        else:
            raise LetOrderValidateException(
                f"Order {self.id} side {self.size} is invalid"
            )

    def place(
        self,
        at: pd.Timestamp,
        raw: object | None = None,
    ) -> "OrderResult":
        """Place `Order`
        Set `status` to `OrderState.Placed`.
        Send event to `Exchange`

        Raises:
            RuntimeError: _description_

        Returns:
            OrderResult: result of `Order`
        """
        self.validate()

        if self.state != OrderState.Pending:
            raise RuntimeError(f"Order {self.id} state {self.state} is not Pending")

        self.state = OrderState.Placed
        self.placed_at = at

        logger.info("Placing new order: %s", self)

        self.exchange.on_order(self)
        return OrderResultOk(order=self, raw=raw)

    def update(
        self,
        limit_price: float | None = None,
        stop_price: float | None = None,
        sl: float | None = None,
        tp: float | None = None,
        raw: object | None = None,
    ) -> "OrderResult":
        """Update Order

        Args:
            limit_price (float, optional): _description_. Defaults to None.
            stop_price (float, optional): _description_. Defaults to None.
            sl (float, optional): _description_. Defaults to None.
            tp (float, optional): _description_. Defaults to None.

        Raises:
            RuntimeError: _description_
        """
        if self.is_closed:
            raise RuntimeError(f"Update a closed order {self}")

        # TODO: validate parameters
        if limit_price is not None:
            self.limit_price = limit_price
        if stop_price is not None:
            self.stop_price = stop_price

        if sl is not None:
            self.sl_price = sl
        if tp is not None:
            self.tp_price = tp

        self.exchange.on_order(self)
        return OrderResultOk(order=self, raw=raw)

    def fill(
        self,
        price: float,
        at: pd.Timestamp,
        raw: object | None = None,
    ) -> "OrderResult":
        """Fill `Order`.
        Set `status` to `OrderState.Executed`.
        Send event to `Exchange`

        Args:
            price (float): Executed price
            at (pd.Timestamp): Executed bar

        Raises:
            RuntimeError: _description_

        Returns:
            OrderResult: result of `Order`
        """
        if self.state != OrderState.Placed:
            raise RuntimeError(f"Order {self.id} state {self.state} is not Placed")

        self.filled_at = at
        self.filled_price = price
        self.state = OrderState.Filled
        self.exchange.on_order(self)
        return OrderResultOk(order=self, raw=raw)

    def cancel(self, raw: object | None = None, **kwargs) -> "OrderResult":
        """Cancel `Order`
        Set `status` to `OrderState.Canceled`.
        Send event to `Exchange`

        Raises:
            RuntimeError: Validate state is `OrderState.Placed`

        Returns:
            OrderResult: result of `Order`
        """
        if self.state != OrderState.Placed:
            raise RuntimeError(f"Order {self.id} state {self.state} is not Placed")

        self.state = OrderState.Canceled
        self.exchange.on_order(self)
        return OrderResultOk(order=self, raw=raw)

    def merge(self, other: "Order"):
        """Update current `Order` variables by other `Order`

        Args:
            other (Order): Merge source `Order`

        Raises:
            RuntimeError: Validate same id
        """
        if other is self:
            return

        if self.id != other.id:
            raise RuntimeError(f"Merge difference id {self.id} != {other.id} order")

        self.size = other.size
        self.sl_price = other.sl_price
        self.tp_price = other.tp_price

        if other.limit_price:
            self.limit_price = other.limit_price
        if other.stop_price:
            self.stop_price = other.stop_price
        if other.placed_at:
            self.placed_at = other.placed_at

        if other.filled_price:
            self.filled_price = other.filled_price
        if other.filled_at:
            self.filled_at = other.filled_at

        if other.parent:
            self.parent = other.parent

    # Fields getters
    @property
    def place_price(self) -> float | None:
        """Getter of place_price

        Returns:
            float | None: `float` or `None`
        """
        if self.type == OrderType.Limit:
            return self.limit_price
        if self.type == OrderType.Stop:
            return self.stop_price
        return None

    @property
    def limit(self) -> float | None:
        """Getter of limit_price

        Returns:
            float | None: `float` or `None`
        """
        return self.limit_price

    @property
    def stop(self) -> float | None:
        """Getter of stop_price

        Returns:
            float | None: `float` or `None`
        """
        return self.stop_price

    @property
    def sl(self) -> float | None:
        """Getter of sl_price

        Returns:
            float | None: `float` or `None`
        """
        return self.sl_price

    @property
    def tp(self) -> float | None:
        """Getter of tp_price

        Returns:
            float | None: `float` or `None`
        """
        return self.tp_price

    # Extra properties
    @property
    def is_sl_order(self) -> bool:
        """`Order` is stop-loss order of a `Position`

        Returns:
            bool: _description_
        """
        return self.parent and self is self.parent.sl_order

    @property
    def is_tp_order(self) -> bool:
        """`Order` is take-profit order of a `Position`

        Returns:
            bool: _description_
        """
        return self.parent and self is self.parent.tp_order

    @property
    def is_opening(self) -> bool:
        """Flag to check `Order` still alive

        Returns:
            bool: True if `state` in [OrderState.Pending, OrderState.Placed, OrderState.Partial]
        """
        return self.state in [OrderState.Pending, OrderState.Placed, OrderState.Partial]

    @property
    def is_closed(self) -> bool:
        """Flag to check `Order` closed

        Returns:
            bool: True if `state` in [OrderState.Filled, OrderState.Canceled]
        """
        return self.state in [OrderState.Filled, OrderState.Canceled]


class OrderResult:
    """Result of `Order`"""

    def __init__(
        self,
        ok: bool = True,
        order: "Order | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            ok (bool | None, optional): Flag to check `Order` is success or not. Defaults to True.
            order (Order | None, optional): Order own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Order`. Defaults to None.
        """
        self.ok: bool = ok
        self.order: "Order | None" = order
        self.raw: object | None = raw

    def _repr_params(self):
        params = f"ok={self.ok} order={self.order}"
        if self.raw is not None:
            params += f"raw='{self.raw}'"
        return params

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._repr_params()}>"


class OrderResultOk(OrderResult):
    """Result of a success `Order`"""

    def __init__(
        self,
        order: "Order | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            order (Order | None, optional): Order own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Order`. Defaults to None.
        """
        super().__init__(ok=True, order=order, raw=raw)


class OrderResultError(OrderResult):
    """Result of a error `Order`"""

    def __init__(
        self,
        error: str,
        order: "Order | None" = None,
        raw: object | None = None,
    ) -> None:
        """_summary_

        Args:
            error (str): Error message
            order (Order | None, optional): Order own the result. Defaults to None.
            raw (object | None, optional): Raw object of `Order`. Defaults to None.
        """
        super().__init__(ok=False, order=order, raw=raw)
        self.error: str = error

    def _repr_params(self):
        return f"error={self.error} {super()._repr_params()}"
